/*****************************************************************************
 * $Id: gfx_math.hh,v 1.51 2006/10/21 00:48:19 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	Math functions for computer graphics.
 * @remarks	To speed or improve FP execution:
 *          - use multiplication instead of division
 *          - power of 2 operations involve no loss of accuracy
 *          - gfx::PI is preferred over M_PI etc from math.h.
 *            math.h has confusing #defines such as M_2_PI
 *            which means 2/PI (not 2*PI).
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#ifndef GFX_MATH_HH
#define GFX_MATH_HH 1

#if GFXSYS_OPENGL
#include "gfx_opengl.hh"
#endif
#include "gfx_stream.hh"

namespace gfx {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  M A C R O S  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// A NaN never equals anything -- not even another NaN.
// C99 isfinite() is true if neither NaN nor infinity.
//#define IFNAN(N) (!((N)==(N))) // (compiler might optimize this incorrectly)
#if __GNUC__  &&  defined(__USE_ISOC99)
#define IFNAN(N) (UX(!isfinite(N)))
#else
// Tested on gcc 3.4.3 x86.
template<typename FP> bool IFNAN( volatile FP& n ) { return !(n==n); }
#endif

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  T R I G   /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// fpx is used in trig functions since they're used in matrix functions
// which need to maintain precision.

// 180' = PI
// 360' = 2 * PI
const fpx PI  = 3.14159265358979323846;
const fpx PI2 = 6.28318530717958647692;   // pi*2

#if SPEED
typedef fp Radian;  // aren't distinct C++ types (useless for overloading functions)
typedef fp Degree;
#else
typedef fpx Radian;
typedef fpx Degree;
#endif

const Radian RADIAN_180 = gfx::PI;   // 180' in radians (half circle)
const Radian RADIAN_360 = gfx::PI2;  // 360' in radians (full circle)

// Compare degrees/radians while ignoring sign.
#define IF_DEG_GT( DEG1, DEG2 ) (ABS(DEG1) > ABS(DEG2))
#define IF_DEG_LT( DEG1, DEG2 ) (ABS(DEG1) < ABS(DEG2))
#define IF_RAD_GT( RAD1, RAD2 ) (ABS(RAD1) > ABS(RAD2))
#define IF_RAD_LT( RAD1, RAD2 ) (ABS(RAD1) < ABS(RAD2))

/*****************************************************************************
 * sincos().
 *****************************************************************************/
template<typename FP>
void
SinCos( Radian rad, FP& si, FP& co )
{
	// This x87 asm code in this C++ template returning args by reference
	// seems perilous but was tested by gfx/tests/math.cc.
#if defined(__GNUC__) && defined(CPU_X86)
	asm("fsincos" : "=t" (co), "=u" (si) : "0" (rad));
#elif defined(__GNUC__) && !defined(__APPLE_CC__)
	sincos( rad, &si, &co );  // GNU extension
#else
	si = sin( rad );
	co = cos( rad );
#endif
}

/*****************************************************************************
 * Truncate a degree.
 *****************************************************************************/
static inline Degree
TruncateDeg( Degree deg )
{
	if ( EX(deg < 360.0) )
		return deg;
	else
		return deg - 360.0;
}

/*****************************************************************************
 * Convert a (360) degree into a radian.
 * 360' = 2 * PI
 *****************************************************************************/
static const fpx degradRatio = (1.0 / 360.0) * (gfx::PI2);
#define DEG2RAD( DEG ) ((DEG) * (degradRatio))
static inline Radian
deg2rad( Degree deg )
{
	return DEG2RAD( deg );
	//return deg / 360.0 * (2.0 * gfx::PI); // x87 FDIV is slower than FMUL
}

/*****************************************************************************
 * Convert a radian into a (360) degree.
 * 360' = 2 * PI
 *****************************************************************************/
static const fpx raddegRatio = (1.0 / gfx::PI2) * 360.0;
#define RAD2DEG( RAD ) ((RAD) * (raddegRatio))
static inline Degree
rad2deg( Radian rad )
{
	return RAD2DEG( rad );
	//return rad / (2.0 * gfx::PI) * 360.0;	// x87 FDIV is slower than FMUL
}

/*****************************************************************************
 * 360' sin().
 *****************************************************************************/
static inline fpx
sin_deg( Degree deg )
{
	return sin( deg2rad( deg ) );
}

/*****************************************************************************
 * 360' cos().
 *****************************************************************************/
static inline fpx
cos_deg( Degree deg )
{
	return cos( deg2rad( deg ) );
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  M A T R I X   ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

#include "gfx_math_matrix.hh"

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  V E C T O R   //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Distance between vector and implied zero origin.
 *****************************************************************************/
static inline fp
Distance( const Vector2& v )
{
	return sqrtf( v.x * v.x
                + v.y * v.y );
}
static inline fp
Distance( const Vector3& v )
{
	return sqrtf( v.x * v.x
                + v.y * v.y
                + v.z * v.z );
}

/*****************************************************************************
 * Distance between two vectors -- DistanceSquared() is faster than this.
 *****************************************************************************/
template<typename VECTOR>  // 2D/3D
fp
Distance( const VECTOR& v1, const VECTOR& v2 )
{
	VECTOR delta = v2 - v1;
	return Distance( delta );
}

/*****************************************************************************
 * Calculate the distance^2 (squared) between two vectors.
 * DISTANCE RETURNED IS SQUARED (distance^2).
 * SQUARE ROOT OF DISTANCE IS SKIPPED FOR SPEED.
 * sqrt() and the underlying FPU instruction is very slow.
 *****************************************************************************/
static inline fpx
DistanceSquared( const Vector3& v1, const Vector3& v2 )
{
	fpx xd = v2.x - v1.x;  // extra precision
	fpx yd = v2.y - v1.y;
	fpx zd = v2.z - v1.z;
	return xd*xd + yd*yd + zd*zd;
}
// Distance^2 between two matrixs.
// Don't pass different kinds of matrixs (eg Eye origin is negative but Dyna is positive).
template<typename MATRIX1, typename MATRIX2>
fpx
DistanceSquaredMatrix( const MATRIX1& m1, const MATRIX2& m2 )
{
	Vector3 v1( m1[Ox], m1[Oy], m1[Oz] );
	Vector3 v2( m2[Ox], m2[Oy], m2[Oz] );
	return DistanceSquared( v1, v2 );
}

/*****************************************************************************
 * @return A normalized vector (unit-length 1.0) of the same vector.
 * Not to be confused with calculating the normal vector of two vectors (CrossProduct).
 *****************************************************************************/
static inline Vector3
Normalize( const Vector3& v )
{
    fp dist = Distance( v );
    return Vector3( v.x / dist,
                    v.y / dist,
                    v.z / dist );
}

/*****************************************************************************
 * Calculate a cross product of two vectors which yields a normal vector
 * that is perpendicular to the plane containing the two source vectors.
 * Note: The cross product won't be unit length.
 * c = a x b
 *****************************************************************************/
static inline void
CrossProduct( NormalVertex& n, /*OUT*/
              const Vector3& v1, /* origin of normal vector */
              const Vector3& v2,
              const Vector3& v3 )
{
	fp a, b, c;
	fp d, e, f;

	a = v2.x - v1.x;
	b = v2.y - v1.y;
	c = v2.z - v1.z;

	d = v3.x - v1.x;
	e = v3.y - v1.y;
	f = v3.z - v1.z;

	n.x = b*f - c*e;
	n.y = c*d - a*f;
	n.z = a*e - b*d;
}

/*****************************************************************************
 * Calculate the dot product of two vectors which yields a scalar.
 * If the angle between the two vectors is between 0 and 90 degrees,
 * the dot product will be positive, else negative.
 *****************************************************************************/
static inline fp
DotProduct( const Vector2& v1, const Vector2& v2 )
{
	return   v1.x * v2.x
           + v1.y * v2.y;
}
static inline fp
DotProduct( const Vector3& v1, const Vector3& v2 )
{
	return   v1.x * v2.x
           + v1.y * v2.y
           + v1.z * v2.z;
}

/*****************************************************************************
 * Calculate angle (in radians) between two vectors sharing the same origin.
 *****************************************************************************/
template<typename VECTOR>  // 2D/3D template
Radian
Angle( const VECTOR& u, const VECTOR& v )
{
	// Dot product formula indirectly provides the angle.
	// angle = arccos( DotProduct(u,v) / Distance(u) * Distance(v) )
	fp uv = Distance(u) * Distance(v);
	if ( ABS(uv) < 0.1 ) return 0.0;  // prevent NaN and div by 0
	fp theta = DotProduct(u,v) / uv;
	theta = Clamp( theta, -1.0, 1.0 ); // acos() won't tolerate 1.00000001
DEBUG_CODE( errno = 0; )
	fp rad = acos( theta );
ASSERT(errno == 0);  // if acos() failed
	return rad;
}

/*****************************************************************************
 * Interpolate between two vectors.
 * @param   v1, v2
 * @param   frac
 *          Fraction in range {0,..,1.0}.
 *          If fraction is 0.0, v1 is returned.
 *          If fraction is 1.0, v2 is returned.
 *          If fraction is 0.5, midpoint is returned.
 * @return Interpolated 3D point.
 *****************************************************************************/
static inline Vector3
Interpolate( const Vector3& v1, const Vector3& v2, const fp fraction )
{
ASSERT( fraction >= 0.0 && fraction <= 1.0 );
    return Vector3( v1 + ((v2-v1) * fraction) );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  M I S C   M A T H   //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * x^2
 *****************************************************************************/
template<typename T>
T
SQUARE( T x )
{
	return x * x;
}

/*****************************************************************************
 * Average.
 *****************************************************************************/
template<typename T>
T
AVG( T x, T y )
{
	return (x + y) / 2;
}

/*****************************************************************************
 * Force a number to be within a range.
 *****************************************************************************/
template<typename T>
T
Range( T val, T lo, T hi )
{
	if ( val < lo ) return lo;
	if ( val > hi ) return hi;
	return val;
}

/*****************************************************************************
 * Map a number from one range into another.
 *****************************************************************************/
static inline fp
Remap( fp value, fp rangeOld, fp rangeNew )
{
	return value/rangeOld*rangeNew;
}

/*****************************************************************************
 * Truncate a float at an interval.
 * Eg, Truncate(24,10) returns 20.
 *****************************************************************************/
static inline fp
Truncate( fp val, fp interval )
{
	float integral;						// not fp as address is passed to modff()
	modff( val/interval, &integral );
	return integral*interval;
}

/*****************************************************************************
 * Normalize a 3D point (unit 1 distance) then return one normalized coordinate.
 * The return value typically is passed to asin().
 *****************************************************************************/
static inline fp
NormalizeCoord( uint axis, const Vector3& v )
{
	fp dist = Distance( v );
	return v[axis] / dist;
}

} // namespace gfx

#endif // GFX_MATH_HH
