/*****************************************************************************
 * $Id: transform_x86_sse.cc,v 1.1 2006/10/21 00:19:08 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	Test speed of x87 vs. SSE vertex transformations.
 * @verbatim
 *			This test proves why x86 SSE isn't used -- SSE is slower than x87!
 *          1/4 of the calculations done by SSE are unnecessary since
 *          the 1/4 of the elements in the 4x4 matrix are unused.
 *          A 4x3 matrix wouldn't help because then SSE would need
 *          a 3-way pipe instead of 4-way to fit.
 *
 *			Will dump core if compiled by gcc-4.0.0.
 *
 * Results on an Intel Pentium III 533/133 (Katmai):
 * /opt/gcc-3.4.3/bin/g++ -O2 -msse -o transform_x86_sse transform_x86_sse.cc
 *
 * Vertex transformation test: 
 * vertex eye (x87) = -365556, -839.292, -348387
 * vertex eye (SSE) = -365556, -839.293, -348387
 * Calculating 50000000 transformations using x87... 
 * Calculating 50000000 transformations using SSE... 
 * time x87: 
 * secs (user+sys) = 3.55
 * secs (user)     = 3.55
 * time SSE: 
 * secs (user+sys) = 4.7
 * secs (user)     = 4.69
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#define TESTING 1
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
using namespace std;

#include "transform_x86_sse.hh"

/*****************************************************************************
 * 
 *****************************************************************************/
void
Transform_x87( EyeVertex& v2, const WorldVertex& v )
{
	fp x = v.x + m[Ox];
	fp y = v.y + m[Oy];
	fp z = v.z + m[Oz];
	v2.x = x * m[Xx] + y * m[Xy] + z * m[Xz];
	v2.y = x * m[Yx] + y * m[Yy] + z * m[Yz];
	v2.z = x * m[Zx] + y * m[Zy] + z * m[Zz];
}

/*****************************************************************************
 * 
 *****************************************************************************/
void
Transform_sse( EyeVertex* ve, const WorldVertex* vw )
{
	fp4 trans, x2, y2, z2;
	trans.v = __builtin_ia32_addps( FP_FP4(vw), FP_FP4(&m[Ox]) );
	x2.v    = __builtin_ia32_mulps( trans.v, FP_FP4(&m[Xx]) );
	y2.v    = __builtin_ia32_mulps( trans.v, FP_FP4(&m[Yx]) );
	z2.v    = __builtin_ia32_mulps( trans.v, FP_FP4(&m[Zx]) );
	ve->x = x2.f[0] + x2.f[1] + x2.f[2];
	ve->y = y2.f[0] + y2.f[1] + y2.f[2];
	ve->z = z2.f[0] + z2.f[1] + z2.f[2];
}

/*****************************************************************************
 * 
 *****************************************************************************/
bool
TestTransformation( void )
{
	// Initialize matrix.
	m[0] = 0.997404;	m[1] = 0.0522642;	m[2] = -0.0495289;	m[3] = 0.0;
	m[4] = -0.0495289;	m[5] = 0.997261;	m[6] = 0.0549317;	m[7] = 0.0;
	m[8] = 0.0522642;	m[9] = -0.052336;	m[10] = 0.997261;	m[11] = 0.0;
	m[12] = -382897;	m[13] = -2165.33;	m[14] = -330162;	m[15] = 0.0;

	WorldVertex vw;
	vw.x = 123.0;
	vw.y = 456.0;
	vw.z = 789.0;

	// Test transformation using x87 code.
	EyeVertex ve_x87;
	Transform_x87( ve_x87, vw );

	// Test transformation using SSE code.
	EyeVertex ve_sse;
	Transform_sse( &ve_sse, &vw );

	cout << "Vertex transformation test: " << endl;
	cout << "vertex eye (x87) = " << ve_x87.x << ", " << ve_x87.y << ", " << ve_x87.z << endl;
	cout << "vertex eye (SSE) = " << ve_sse.x << ", " << ve_sse.y << ", " << ve_sse.z << endl;

	return true;
}

/*****************************************************************************
 * 
 *****************************************************************************/
bool
TestTransformationSpeed( void )
{
	// Initialize matrix.
	m[0] = 0.997404;	m[1] = 0.0522642;	m[2] = -0.0495289;	m[3] = 0.0;
	m[4] = -0.0495289;	m[5] = 0.997261;	m[6] = 0.0549317;	m[7] = 0.0;
	m[8] = 0.0522642;	m[9] = -0.052336;	m[10] = 0.997261;	m[11] = 0.0;
	m[12] = -382897;	m[13] = -2165.33;	m[14] = -330162;	m[15] = 0.0;

	EyeVertex ve_x87, ve_sse;
	WorldVertex vw;
	vw.x = 123.0;
	vw.y = 456.0;
	vw.z = 789.0;

	const int cnt = 50 *1000 * 1000;

	// POSIX.
	struct tms time1_x87, time2_x87;
	struct tms time1_sse, time2_sse;

	// Test transformation using x87 code.
	cout << "Calculating " << cnt << " transformations using x87... " << endl;
	times( &time1_x87 );
	for ( int i = 0; i < cnt; ++i )
		Transform_x87( ve_x87, vw );
	times( &time2_x87 );

	// Test transformation using SSE code.
	cout << "Calculating " << cnt << " transformations using SSE... " << endl;
	times( &time1_sse );
	for ( int i = 0; i < cnt; ++i )
		Transform_sse( &ve_sse, &vw );
	times( &time2_sse );

	cout << "time x87: " << endl; PrintTimeDelta( time1_x87, time2_x87 );
	cout << "time SSE: " << endl; PrintTimeDelta( time1_sse, time2_sse );

	return true;
}

/*****************************************************************************
 * 
 *****************************************************************************/
int main()
{
	TestTransformation();
	TestTransformationSpeed();
}
